/*

samtoms.c

type samtoms for usage instructions

I've been using Borland C++ for windows (it's the only C compiler
I can use on DOS :)) and it compiles fine with Tiny memory model.

Make sure you compile with byte-alignment though!

Doesn't work with win95, you can't use biosdisk:)

My executable is 20ish k. I use size optimisation (global, all
possible size optimisations) since most of your time is sat waiting
for the disk...

*/

#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <bios.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

typedef struct {
	unsigned char filetype; /* 0 */
	unsigned char filename[10]; /* 1 */
	unsigned char msb_numsecs;  /* 11 */
	unsigned char lsb_numsecs;  /* 12 */
	unsigned char start_track;  /* 13 */
	unsigned char start_sec;    /* 14 */
	unsigned char SAM[195];     /* 15 */
	unsigned char fut_and_past[10]; /*210*/
#define diskname fut_and_past
	/* just to make it obvious */
	unsigned char flags; /* 220 */
	unsigned char type_info[11]; /* 221 */
	unsigned char reserved[4]; /* 232 */
	unsigned char start_page; /* 236 */
	unsigned int start_offset; /* 237 */
	unsigned char length_page; /* 239 */
	unsigned int length_modulo; /* 240 */
	unsigned char exec_page; /* 242 */
	unsigned int exec_modulo; /* 243 */
	unsigned char day; /* 245 */
	unsigned char month; /* 246 */
	unsigned char year; /* 247 */
	unsigned char hour; /* 248 */
	unsigned char minute; /* 249 */
	unsigned char spare[4]; /* 250 */
	unsigned int subdir; /* 254 */
	/*total 255 */

	/* OK, so this is technically incorrect, but it works for my purposes.
		I'll update it when I can be bothered to find out what the exact
		entry ought to be... or when it's proved necessary, whichever is
		the later ;-) */
} dir_entry;

/* stuff for disk formatting */

#define SECTORS_PER_TRACK 10

/* this bit from Eduard Dong's Dos formatter */
void interrupt (*System_dpb)();

struct DPB
		  {
		  char spec1;             /* first specify byte */
		  char spec2;                     /* second specify byte */
		  char stop_delay;        /* delay until motor turned off (ticks) */
		  char bps;               /* bytes per sector 0=128, 1=256, 2=512, 3=1024 */
		  char spt;               /* sectors per track */
		  char gap;               /* gap between sectors (2Ah for 5.25", 1Bh for 3.5") */
		  char data_l;            /* data length (ignored if bps is non-zero) */
		  char gap_l;             /* formatting gap length (50h for 5.25", 6ch for 3.5") */
		  char filler;            /* format filler byte (default F6h) */
		  char settle;            /* head settle time in milliseconds */
		  char start_time;        /* motor start time in 1/8 seconds */
		  } my_dpb;

void
reset_disk_controller (void) {
	setvect(0x1e, System_dpb);
	biosdisk(_DISK_RESET, 0, 0, 0, 0, 0, (void *)0);
	return;
}

int kdh; /* use kedisk headers to save? */

unsigned char *
makename (char *getfrom) {

	static char buff[11];
	int x;

	strcpy(buff, "          ");
	for (x=0; (x < 10) && (getfrom[x]); x++) {
		(buff[x]=getfrom[x]);
	}
	buff[10]=0;
	return buff;
}

void
tryit (int bioscommand, char *buffer, int maxtries, int track,
		 int sector, int numsecs) {

	static char *diskerrs[]={ /* 0=no error, 1=bad command (can't happen) */
	"", "", "Address mark not found",
	"", "Sector not found",/*3 is write protect, 5 is HD only */
	"", "Disk changed", "Drive parameter activity failed",
	"DMA overrun", "Attempt to perform DMA across 64k boundary",
	"Bad sector", "Bad track", "Unsupported track", "Unknown", "Status error"};
	int status, tries, head, x;

	if (track > 127) {
		track-=128;
		head=1;
	} else {
		head=0;
	}
	for (status=1, tries=0; (tries < maxtries) && status; tries++) {
		status=biosdisk(bioscommand, 0, head, track, sector, numsecs, buffer);
		if (status) {
			switch (status) {
				case 0x03:
					puts("Disk is write protected!");
					exit(1);
				case 0x10:
					printf("Bad CRC/ECC on disk read at %d, %d, retrying\n", track+(head<<7),sector);
					break;
				case 0x11:
					printf("Warning -- CRC/ECC corrected data error at %d, %d.\n", track+(head<<7),sector);
					status=0;
					break;
				case 0x20:
					puts("Controller failed");
					tries=maxtries;
					break;
				case 0x40:
					printf("Seek failed at %d, %d, retrying\n", track+(head<<7), sector);
					break;
				case 0x80:
					puts("Is there a disk in drive 0?");
					tries=maxtries;
					break;
				case 0xcc:
					printf("Write failed at %d, %d, retrying\n", track+(head<<7), sector);
					break;
				case 0xff:
					printf("Sense operation failed at %d, %d, retrying\n", track+(head<<7), sector);
					break;
				default:
					printf("%s at %d, %d, retrying\n", diskerrs[status], track+(head<<7), sector);
					break;
			}
		}
	}
	if (status) {
		if (numsecs > 1) {
			/* find which sector it really is... */
			for (x=0; x < numsecs; x++) {
				status=biosdisk(bioscommand, 0, head, track, sector+x, 1, buffer+(x*512));
				if (status) break;
			}
			printf("Disk error accurately at %d, %d.\n", track+(head<<7), sector+x);
		}
		puts("Fatal");
		exit(1);
	}
}
void
sam_form(void) {

	int track, head, sec, status;

	struct {
		unsigned char cyl;
		unsigned char head;
		unsigned char sec;
		unsigned char num;
	} format_header[SECTORS_PER_TRACK];

/*      Get and save address of old disk parameter block, copy to a replacement
		  block (fields will be changed later).  Restore the old block when we
		  terminate.
 */
	System_dpb = getvect(0x1e);
	movedata(FP_SEG(System_dpb), FP_OFF(System_dpb),FP_SEG(&my_dpb),
	  FP_OFF(&my_dpb), sizeof(struct DPB));
	/* copies the info from the system's dpb to our own dpb struct */

	setvect(0x1e, (void interrupt (*)()) &my_dpb);
	atexit(reset_disk_controller);

	my_dpb.spt=SECTORS_PER_TRACK;
	my_dpb.bps=2; /* 512 bytes per sector */
	my_dpb.gap=0x1b; /* 3.5 disk */
	my_dpb.gap_l = 0x6c;
	my_dpb.filler=0; /* clear sectors to start */
	poke (0x40, 0x90, 0x54); /* what the ---- does this do? */

	status=biosdisk(0x18, 0, 0, 80, SECTORS_PER_TRACK, 0, format_header);

	if (status & 0xff00u) {
		printf("Status error %d on ram controller (biosdisk 0x18)\n", status);
		exit(1);
	}
	status=biosdisk(0x17, 0, 0, 0, 0, 4, format_header);
	if (status) {
		printf("Status error %d on drive calibration (biosdisk 0x17)\n", status);
	}

	for (track=0; track < 80; track ++) {
		for (head=0; head < 2; head ++) {
			for (sec=0; sec < SECTORS_PER_TRACK; sec++) {
				format_header[sec].cyl=track;
				format_header[sec].head=head;
				format_header[sec].sec=sec+1;
				format_header[sec].num=2; /* bytes/sector -- 0=128, 1=256, 2=512... */
			}
			tryit(_DISK_FORMAT, (char *)&format_header[0], 4, track+128*head, 0, SECTORS_PER_TRACK);
			printf("Successfully formatted track \033[32m%d\033[37m head \033[32m%d\033[37m\015", track, head);
		}
	}
	reset_disk_controller();
}

int
create_sam (unsigned char *create, unsigned char *already, int numsecs) {

/* create Sector Address Map
	already[] contains the BAM for the disk, create[] is where we're
	going to put the damn things */

	int sec;

	for (sec=0; (sec < 1560) && (numsecs); sec++) {
		if (((already[sec>>3]) & (1<<(sec&7)))==0) {
			create[sec>>3]|=1<<(sec&7);
			numsecs--;
		}
	}
	return(numsecs);
}

int
write_coup (dir_entry *direc, FILE *infile) {

/* completely ignores length - just works on what it gets from the file.
	this is how it should be since all the directory stuff is done outside the file */

	int oldsec, sec, x;
	int bread, wrtrack;

	char buffer[512];

	oldsec=0;

	for (sec=0; sec < 1560; sec++) {
		if (direc->SAM[sec>>3] & (1 << (sec & 7))) {
			if (oldsec) {
				wrtrack=sec / 10 + 4;
				if (wrtrack > 79) {
					wrtrack+=48;
				}
				buffer[511]=(sec%10)+1;
				buffer[510]=wrtrack;
				wrtrack=oldsec / 10 + 4;
				if (wrtrack > 79) {
					wrtrack+=48;
				}
				tryit(_DISK_WRITE, buffer, 4, wrtrack, (oldsec%10)+1, 1);
				bread=fread(buffer, sizeof(unsigned char), 510, infile);
				if (bread < 510) {
					for (x=bread; x < 510; x++) buffer[x]=0;
				}
			} else {
				direc->start_sec=(sec%10)+1;
				wrtrack=(sec/10)+4;
				if (wrtrack>79) wrtrack+=48;
				direc->start_track=wrtrack;

				/* writes the "first sector" information to the directory entry */

				/* now copy the info like file length onto the file header */

				buffer[0]=direc->filetype;
				*((int *)(buffer+1))=direc->length_modulo;
				*((int *)(buffer+3))=direc->start_offset;
				buffer[5]=0;             /* "Unused"... */
				buffer[6]=0;
				buffer[7]=direc->length_page;
				buffer[8]=direc->start_page;

				bread=fread(buffer+9, sizeof(unsigned char), 501, infile);
				if (bread < 501) {
					for (x=bread; x < 501; x++) buffer[x]=0;
				}
			}
			oldsec=sec;
		}
	}
	buffer[510]=0;
	buffer[511]=0;
	wrtrack=oldsec / 10 + 4;
	if (wrtrack > 79) {
		wrtrack+=48;
	}
	tryit(_DISK_WRITE, buffer, 4, wrtrack, (oldsec%10)+1, 1);
	return (0);
}

void
read_kedisk_header (FILE *infile, dir_entry *direc) {
/* this is for writing KDSK files back from MS-Dos format.

	direc should be the location of the new directory entry that
	is being formed. infile is obvious. */

	fread (direc, 1, 11, infile);
	/* type and sam name */
	fread (direc->type_info, 1, 11, infile);
	/* file type information */
	fread (&(direc->start_page), 1, 14, infile);
	/* other info - length etc, and date */
}

void
write_kedisk_header (FILE *outfile, dir_entry *direc) {

/* writes the header out to the DOS file.

	direc points to the directory entry from the disk */

	fputs("KDSK", outfile);
	/* magic number */
	fwrite(direc, 1, 11, outfile);
	/* file type and sam filename */
	fwrite(direc->type_info, 1, 11, outfile);
	/* file type information */
	fwrite (&(direc->start_offset), 1, 14, outfile);
	/* 236-250

	236                      - start page number
	237 & 238<<8 - page offset
	239              - number of pages in length
	240 & 241<<8 - length modulo 16384
	242 & 243<<8 - execution address (or LINE number if)
	244          - execution page    (BASIC file       )
	245                 - day
	246                      - month
	247                      - year
	248          - hour
	249                      - minute
*/

/* 40 bytes in all. Not too bad, I'd say. Is it worth having
	the date aswell??? */
}


void
readcoup(dir_entry *entry, char *filename) {

	FILE *saveit;
	unsigned long length;
	int track, sector;
	static unsigned char secbuf[512];

	track=entry->start_track;
	sector=entry->start_sec;
	length=entry->length_modulo|((unsigned long)(entry->length_page)<<14);
	printf("File %s length %ld\n", filename, length);

	saveit=fopen(filename, "wb");

	/* w - write, b - BINARY -- DON'T FORGET IT!!!!
	  (bloody DOS...) */

	if (kdh) write_kedisk_header(saveit, entry);

	tryit(_DISK_READ, secbuf, 4, track, sector, 1);

	if (length > 501) {
		fwrite (secbuf+9, 1, 501, saveit);
		length-=501;
		track=secbuf[510];
		sector=secbuf[511];
		do {
			tryit(_DISK_READ, secbuf, 4, track, sector, 1);
			if (length > 510) {
				fwrite(secbuf, 1, 510, saveit);
				length-=510;
			} else {
				fwrite (secbuf, 1, (int) length, saveit);
				length=0;
			}
			track=secbuf[510];
			sector=secbuf[511];
		} while (sector && (length > 0));
		if (length > 0) printf("File ended prematurely: length remains at %ld\n", length);
		if (sector) printf("Length hit 0 but file continues to sector %d, %d...\n", track, sector);
	} else {
		fwrite (secbuf+9, 1, (unsigned int)length, saveit);
		length=0;
	}
	fclose(saveit);
}

void
getdir (int thistrack, dir_entry *buffd) {

	tryit (_DISK_READ, (char *)buffd, 4, thistrack, 1, 10);
	if ((thistrack==0) && (buffd->subdir==0xAA55)) {
		puts("This is an MSDOS format disk");
		exit (1);
	}
}


int
main(int argc, char **argv) {

	dir_entry *builddir;
	unsigned char buildtrack;
	static unsigned char BAM[195]; /* bit address map */
	static struct stat statbuf;

	int done, thisarg, track, sector, strt, command, x, totsecs, c;
	unsigned long length;
	FILE *infile;
	dir_entry *direc;
	unsigned char magic[5];
	char *thisname, tmpch, secbuff[512];

	direc=malloc(sizeof(dir_entry)*20);
	kdh=0;
	command=0; /* read */
	if (argc < 2) {
		puts(
"Usage:\nsamtoms /f | [/d] [/h] [/o t s] {[/r] [/w] [/e] filename}\n\
/h\tUse KDSK headers when storing on MS-Dos\n/r\tRead mode\n\
/w\tWrite mode\n/e\tErase mode\n/d\tDirectory\n\
/o t s\tOutput sector to screen\n\
/f\tFormat disk to Sam format (doesn't work yet)\n\
Note that /r, /w and /e continue up until the next /r, /w or /e\n\n\
Default mode is /r");
		exit (1);
	}
	for (thisarg=1; thisarg < argc; thisarg++) {
		if (argv[thisarg][0]=='/') {
			switch (argv[thisarg][1]) {
				case 'f':
					/* sam disk format -- doesn't allow any commands after... */
					sam_form();
					exit (0);
					break;
				case 'n':
					/* name sam disk */
					thisarg++;
					if (thisarg==argc) { /* no name specified!!! */
						puts("No name specified!!!");
						exit (1);
					}
					tryit(_DISK_READ, (char *)direc, 4, 0, 1, 1);
					strncpy (direc[0].diskname, makename(argv[thisarg]), 10);
					tryit(_DISK_WRITE, (char *)direc, 4, 0, 1, 1);
					break;
				case 'h':
					kdh=1;
					break;
				case 'o':
					/* output sector */
					track=atoi(argv[++thisarg]);
					sector=atoi(argv[++thisarg]);
					tryit(_DISK_READ, secbuff, 4, track, sector, 1);
					fputs("\033[36m000\033[37m ", stdout);
					for (x=0; x < 512; x++) {
						tmpch=secbuff[x];
						if (tmpch < 32) tmpch=' ';
						printf("%02X\033[32m%c\033[37m", secbuff[x], tmpch);
						if (((x < 255) && (x%24==23)) || ((x > 254) && (x%24==15))) {
							printf("\n\033[36m%03d\033[37m ", x+1);
						}
					}
					break;
				case 'r':
					command=0;
					break;
				case 'w':
					command=1;
					break;
				case 'd':
					command=2; /* directory... just simple! */
					c=0;
					totsecs=0;
					for (track=0; (track < 4) && (c < 8); track++) {
						getdir(track, direc);
						if (track==0) {
							printf("\033[36m");
							switch((direc[0].diskname[0]) & 127) {
								case 127:
									puts("Samdos");
									break;
								case '*':
									puts("Masterdos");
									break;
								default:
									puts(makename(direc[0].diskname));
									break;
							}
							printf("\033[37m");
						}
						for (strt=0; (c < 8) && (strt < 20); strt++) {

							if ((direc[strt].filetype | direc[strt].filename[0])==0) {
								c+=8; /* illegal value of c == quit... */
							} else if (direc[strt].filetype) {
								totsecs+=direc[strt].lsb_numsecs | (direc[strt].msb_numsecs<<8);
								fputs(makename(direc[strt].filename), stdout);
								c++;
								if (c==5) {
									fputc('\n', stdout);
									c=0;
								} else {
									fputc('\t', stdout);
								}
							}
						}
					}
					if (c-8) fputc('\n', stdout);
					printf("\033[32m%d\033[37m sectors used. \033[36m%dk\033[37m free.\n", totsecs, (1560-totsecs)>>1);
					break;
				case 'e':
					command=3; /* erase */
					break;
				default:
					printf("Unknown switch %c\n", argv[thisarg][1]);
			}
		} else {
			switch(command) {
				case 1:
					/* write to sam disk!!! */
					memset(BAM, 0, 195);
					done=0;
					buildtrack=255;

					thisname=makename(argv[thisarg]);
					infile=fopen(argv[thisarg], "rb");
					if (infile) {
						for (track=0; (track < 4) && (!done); track++) {
							getdir(track, direc);

							for (strt=0; (done == 0) && (strt < 20); strt++) {
								if ((direc[strt].filetype | direc[strt].filename[0])==0) {
									done=1; /* last entry */
								}
								if (direc[strt].filetype) {
									if (strncmp(direc[strt].filename, thisname, 10)==0) {

									/* this file has the same name -- overwrite it,
										and -don't- include the SAM... */
										printf("Overwriting %s\n", thisname);
										builddir=&direc[strt];
										buildtrack=track;
									} else {
								/* there is a file entry, it's not deleted */
										for (x=0; x < 195; x++) {
											BAM[x]|=direc[strt].SAM[x];
										}
									}
								} else if (buildtrack==255) {
								/* 255 means we haven't yet found a free space */
									buildtrack=track;
									builddir=&direc[strt];
								}
							}
						}
						/* now got BAM */
						getdir(buildtrack, direc);
						/* load in the track that we had the free entry in */

						memset(builddir, 0, 255);
						fread(magic, sizeof (unsigned char), 4, infile);
						stat(argv[thisarg], &statbuf);
						length=statbuf.st_size;

						totsecs=(int)((length+9)>>9);
						/* add 9 bytes for file header, shift 9 for /512 */

						if ((length+9) & 511) totsecs++;
						/* the MOD value */

						builddir->msb_numsecs=totsecs>>8;
						builddir->lsb_numsecs=totsecs&0xff;
						/* bloody things are the wrong way round... thanks Andy... */

						if (strncmp(magic, "KDSK", 4)==0) {
							/* KDSK header file */
							read_kedisk_header(infile, builddir);
						} else {
							/* not headed */
							fseek(infile, 0, SEEK_SET); /* go back to beginning! */
							builddir->filetype=19;      /* code file */
							strncpy(builddir->filename, thisname, 10);
							builddir->start_offset=32768u;
							builddir->start_page=1u; /* save code start 32768 */

							builddir->length_page=(unsigned char) (length>>14);
							builddir->length_modulo=(unsigned int) (length&0xffffu);

							builddir->exec_page=0xffu;
							builddir->exec_modulo=0xffffu;
							/* execution address - what should it be??? */
						}
						builddir->fut_and_past[0]=0;
						builddir->fut_and_past[1]=builddir->filetype;
						*((int *)(builddir->fut_and_past+2))=builddir->length_modulo;
						*((int *)(builddir->fut_and_past+4))=builddir->start_offset;
						builddir->fut_and_past[6]=0;
						builddir->fut_and_past[7]=0;
						builddir->fut_and_past[8]=builddir->length_page;
						builddir->fut_and_past[9]=builddir->start_page;
						builddir->flags=32; /* apparently this should be 32... don't ask me why */
						memset(builddir->spare, 255, 3);
						builddir->subdir=0;
						if (create_sam (builddir->SAM, BAM, totsecs)) {
							printf("No room for file %s\n", argv[thisarg]);
						} else {
							write_coup (builddir, infile);
							tryit(_DISK_WRITE, (char *)direc, 4, buildtrack, 1, 10);
							/* directory entry */
						}
						fclose(infile);
					} else {
						printf("Couldn't open %s for reading\n", argv[thisarg]);
					}
				break;

				case 0:
					done=0;
					for (track=0; (track < 4) && (done==0); track++) {
						getdir(track, direc);
						for (strt=0; (done == 0) && (strt < 20); strt++) {
							if ((direc[strt].filetype | direc[strt].filename[0])==0) {
								printf("File %s not found!\n", argv[thisarg]);
								done=1;
							}
							if (direc[strt].filetype && (strncmp(direc[strt].filename, argv[thisarg], strlen(argv[thisarg]))==0)) {
								readcoup(&direc[strt], argv[thisarg]);
								done=1;
							}
						}
					}
				break;

				case 3:
				/* erase filename.... -must- match exactly */
					done=0;

					thisname=makename(argv[thisarg]);

					for (track=0; (track < 4) && (done==0); track++) {
						getdir(track, direc);
						for (strt=0; (done == 0) && (strt < 20); strt++) {
							if ((direc[strt].filetype | direc[strt].filename[0])==0) {
								printf("File %s not found!\n", thisname);
								done=1;
							}
							if (direc[strt].filetype) {
								if (strncmp(direc[strt].filename, thisname, 10)==0) {
									direc[strt].filetype=0;
									tryit(_DISK_WRITE, (char *)direc, 4, track, 1, 10);
									printf("Erased %s\n", thisname);
									done=1;
								}
							}
						}
					}
				break;

				default:
					printf("Invalid command %d\n", command);
				break;
			}
		}
	}
	return 0;
}
